from uuid import uuid1
from log_request_id import local, NO_REQUEST_ID


class Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def get_request_id():
    try:
        request_id = local.request_id
    except AttributeError:
        request_id = NO_REQUEST_ID if NO_REQUEST_ID else uuid1().hex
    return request_id



# rest-framework 自定义认证

import traceback
from base64 import b64decode
from rest_framework import authentication, exceptions
from ywkd_tools.inner_service_apis import InnerServices
from ywkd_tools.logger import logger


class Auth(object):
    """
    Auth
    """

    service_name = None
    rpc_base_url = None
    rpc_secret = None
    inited = False

    def __init__(self):
        raise TypeError('Cannot create "Auth" instances.')

    @classmethod
    def setup(cls, service_name=None, rpc_base_url=None, rpc_secret=None):
        cls.service_name = service_name
        cls.rpc_base_url = rpc_base_url
        cls.rpc_secret = rpc_secret

        if InnerServices.inited:
            cls.service_name = InnerServices.service_name
            cls.rpc_base_url = InnerServices.base_url
            cls.rpc_secret = InnerServices.secret
        elif service_name and rpc_base_url:
            InnerServices.setup(service_name=service_name,
                                base_url=rpc_base_url,
                                secret=rpc_secret)
        cls.inited = True

    @classmethod
    def inner(cls, authentication_class):
        authentication_class.auth_class = cls
        setattr(cls, authentication_class.__name__, authentication_class)


@Auth.inner
class RPCBasciAuthentication(object):
    """
    RPC 认证 (只能用于RPC)
        headers:
            HTTP_AUTHORIZATION=basci [username]:[password]
    """

    auth_class = None

    def do_check(self, request, username, password):
        if password != self.auth_class.rpc_secret:
            logger.info('密码/用户错误')
            return False
        return True

    def authenticate(self, request):
        if not self.auth_class.inited:
            raise Exception('Auth not setup yet, please do this first!')

        # 没有设置 rpc sercret 则忽略认证
        if not self.auth_class.rpc_secret:
            return True

        username = password = None
        basic_auth = request.META.get('HTTP_AUTHORIZATION')
        if not basic_auth:
            logger.info('缺少 basci authrization')
            return False

        try:
            auth_method, auth_string = basic_auth.split(' ', 1)
        except Exception:
            logger.info('http头部 HTTP_AUTHORIZATION 格式错误')
            return False

        if auth_method.lower() == 'basic':
            try:
                auth_string = b64decode(auth_string.strip())
                username, password = auth_string.decode().split(':', 1)
            except Exception:
                logger.info('错误的 Basic 格式')
                return False
        else:
            logger.info('需要 basic authrization')
            return False

        res = self.do_check(request, username, password)
        if not res:
            return False
        return  True


@Auth.inner
class RSTFBasciAuthentication(authentication.BaseAuthentication):
    """
    Django rest_framework Basci 认证: (只能用于Django rest_framework)
        headers:
            HTTP_AUTHORIZATION=basci [user_id]:[token]
    """

    auth_class = None

    def do_check(self, request, user_id, token):
        service_name = self.auth_class.service_name
        api_name = request.resolver_match.url_name
        operator_type = request.method

        # 检测接口访问权限
        res = None
        try:
            res = InnerServices.Cperm().check_api_operate_perms(
                user_id=user_id,
                token=token,
                service_name=service_name,
                api_name=api_name,
                operator_type=operator_type)
        except Exception as e:
            traceback.print_exc()
            raise exceptions.AuthenticationFailed('与认证服务cperm通信失败')

        if res['code'] != 0:
            raise exceptions.AuthenticationFailed(res['codemsg'])

        # 获取用户信息
        user = InnerServices.Cperm().get_user(user_id)
        user['token'] = request.token
        user['basic_auth'] = request.basic_auth
        return user

    def authenticate(self, request):
        if not self.auth_class.inited:
            raise Exception('Auth not setup yet, please do this first!')

        user_id = token = None
        basic_auth = request.META.get('HTTP_AUTHORIZATION')
        if not basic_auth:
            raise exceptions.AuthenticationFailed('缺少 basci authrization')

        try:
            auth_method, auth_string = basic_auth.split(' ', 1)
        except Exception:
            raise exceptions.AuthenticationFailed('http头部 HTTP_AUTHORIZATION 格式错误')

        if auth_method.lower() == 'basic':
            try:
                auth_string = b64decode(auth_string.strip())
                user_id, token = auth_string.decode().split(':', 1)
            except Exception:
                traceback.print_exc()
                raise exceptions.AuthenticationFailed('错误的 Basic 格式')
        else:
            raise exceptions.AuthenticationFailed('需要 basic authrization')

        request.token = token
        request.basic_auth = basic_auth
        res = self.do_check(request, user_id, token)
        return (res, None)



from ywkd_tools.inner_service_apis.inner_service_apis import InnerServices, InnerService
from ywkd_tools.inner_service_apis.errors import InnerAPIError


@InnerServices.inner
class Participant(InnerService):
    """用户中心"""

    url = '/cperm/rpc/'

    def get_participant(self, participant_type, participant_id):
        """
        获取参与利润分配的公司/团体详情
                @param: participant_type: 类型
                    KTV: ktv
                    VOD: vod
                    代理商: agentibus
                    行业协会(机构): industry_association
                    垫付方: advance_party
                    音乐著作协会: music_copyright_society
                    权力方: copyright_party
                    平台: platform
                @param: id
                @out: {     // 返回内容参考: http://yapi.yuwangkedao.com/project/251/interface/api/6233
                    id: [id],
                    participant_type: 类型,
                    name: 名称,
                    area_code_list: 区域列表,
                    ...
                }
        """
        return self.rpc_client.get_participant(participant_type, participant_id)



import re
import log_request_id
from xmlrpc.client import Transport, SafeTransport, ServerProxy
from base64 import b64encode
from urllib.parse import urljoin
from functools import wraps
from django.http import HttpRequest
from django.conf import settings
from rest_framework.request import Request
from ywkd_tools.inner_service_apis.errors import InnerAPIError
from ywkd_tools.logger import logger
from ywkd_tools.utils import get_request_id


class SpecialTransport(Transport):
    """
    SpecialTransport (http)
    """

    preset_headers = {}

    def set_header(self, key, value):
        if key.lower().startswith('http'):
            key = key[5:]
        if value == None:
            self.preset_headers.pop(key, None)
        else:
            self.preset_headers[key] = value

    def send_headers(self, connection, headers):
        for key, val in self.preset_headers.items():
            connection.putheader(key, val)
        super().send_headers(connection, headers)

    # def send_content(self, connection, request_body):
    #     connection.putheader("Content-Type", "text/xml")
    #     connection.putheader("Content-Length", str(len(request_body)))
    #     connection.endheaders()
    #     if request_body:
    #         connection.send(request_body)


class SpecialSafeTransport(SafeTransport):
    """
    SpecialSafeTransport (https)
    """

    preset_headers = {}

    def set_header(self, key, value):
        if key.lower().startswith('http'):
            key = key[5:]
        if value == None:
            self.preset_headers.pop(key, None)
        else:
            self.preset_headers[key] = value

    def send_headers(self, connection, headers):
        for key, val in self.preset_headers.items():
            connection.putheader(key, val)
        super().send_headers(connection, headers)

    # def send_content(self, connection, request_body):
    #     connection.putheader("Content-Type", "text/xml")
    #     connection.putheader("Content-Length", str(len(request_body)))
    #     connection.endheaders()
    #     if request_body:
    #         connection.send(request_body)


class InnerServices(object):
    """
    InnerServices
    """

    service_name = ''
    base_url = ''
    secret = ''
    inited = False

    def __init__(self):
        raise TypeError('Cannot create "InnerServices" instances.')

    @classmethod
    def setup(cls, service_name, base_url, secret):
        cls.base_url = base_url
        services = [getattr(cls, key) for key in cls.__dict__ if isinstance(getattr(cls, key), InnerService)]
        for service in services:
            service.init(service_name, base_url, secret)
        cls.inited = True

    @classmethod
    def inner(cls, service_cls):
        service = service_cls()
        setattr(cls, service_cls.__name__, service)


class InnerService(object):
    """
    InnerService
    """

    url = None
    client = None
    http_request = None
    request_id = None

    def __init__(self, http_request=None, request_id=None):
        """
            @parms http_request: django.http.HttpRequest 或 rest_framework.request.Request
            @parms request_id: 不指定时 默认使用上下文中的request_id, 建议在Django View 中无特殊需求则无需
                指定 http_request or request_id, 在异步任务中建议指定该值, 保证log日志的完整性
        """
        if not hasattr(self, 'url'):
            raise InnerAPIError('InnerService class "%s" need has a url attr!' %
                                self.__class__.__name__)
        if http_request:
            assert (isinstance(http_request, HttpRequest) or isinstance(http_request, Request)), (
                'The `request` argument must be an instance of '
                '`django.http.HttpRequest` or `rest_framework.request.Request`, not `{}.{}`.'
                .format(http_request.__class__.__module__, http_request.__class__.__name__)
            )
        if http_request:
            self.http_request = http_request
            self.request_id = http_request.id
        if request_id:
            self.request_id = request_id

    def __call__(self, http_request=None, request_id=None):
        new_instance = self.__class__(http_request=http_request, request_id=request_id)
        new_instance.init(self.service_name, self.base_url, self.secret)
        return new_instance

    def get_url(self):
        """获取url"""
        url = self.url
        if not re.match(r'^(http|https)://', self.url, re.I):
            assert (self.base_url != None and re.match(r'^(http|https)://', self.base_url, re.I)), (
                'base_url must not null and start with http or https'
            )
            url = urljoin(self.base_url, self.url)
        return url

    def get_transport(self):
        """获取transport"""
        transport = None
        if self.http_url.startswith('https'):
            transport = SpecialSafeTransport()
        else:
            transport = SpecialTransport()

        # 认证: HTTP_AUTHORIZATION basic xxxxxxxx
        basic_auth_str = '{service_name}:{secret}'.format(**{
            'service_name': self.service_name,
            'secret': self.secret
        })
        header_auth_str = 'basic {}'.format(
            b64encode(bytes(basic_auth_str, 'utf-8')).decode('utf-8'))
        transport.set_header('Authorization', header_auth_str)
        return transport

    def init(self, service_name, base_url, secret):
        # logger.debug('Init inner service: "{}"'.format(self.__class__.__name__))
        self.service_name = service_name
        self.base_url = base_url
        self.secret = secret
        self.http_url = self.get_url()
        self.transport = self.get_transport()
        self.client = ServerProxy(self.http_url, transport=self.transport, allow_none=True, use_datetime=True)

    @property
    def rpc_client(self):
        assert self.client, (
            '"{}" instance not setup yet, please do init first!'
            .format(self.__class__.__name__)
        )

        # 在请求头中添加 request_id
        request_id = None
        if self.request_id:
            request_id = self.request_id
        else:
            request_id = get_request_id()
        # request_id_header = getattr(settings, log_request_id.REQUEST_ID_HEADER_SETTING, 'X-Request-Id')  # bug HTTP_X_REQUEST_ID 不能直接使用
        self.transport.set_header('X-Request-Id', request_id)

        return self.client



from ywkd_tools.inner_service_apis.inner_service_apis import InnerServices, InnerService
from ywkd_tools.inner_service_apis.errors import InnerAPIError


@InnerServices.inner
class MSG(InnerService):
    """消息"""

    url = '/cperm/rpc/'

    def send_msg(self, user_ids, msg, redirect_url=None, ignore_err=True):
        """
        发送站内消息
            @parms user_ids: 接收message 的用户id (可以为list)
            @parms msg: 消息内容
            @parms redirect_url: 跳转链接
            @parms ignore_err: 是否忽略错误
        """
        return self.rpc_client.send_msg(user_ids, msg, redirect_url, ignore_err)

    def send_email(self, user_ids, mail_subject, mail_message, countdown=None, eta=None, ignore_err=True):
        """
        发送email
            @parms user_ids: 接收email 的用户id (可以为list)
            @parms mail_subject: 邮件主题
            @parms mail_message: 邮件内容
            @parms countdown: 是否延迟发送, 延时几秒, 例如: countdown=5
            @parms eta: 是否延迟发送, 延时几秒, 例如: eta=datetime.datetime.utcnow() + datetime.timedelta(seconds=5) 注意时间必须为 utc时间
            @parms ignore_err: 是否忽略错误


        例:
            import datetime
            eta = datetime.datetime.utcnow() + datetime.timedelta(seconds=3)
            InnerServices.MSG().send_email([57], 'Email Test A', 'test', None, eta)
        """
        return self.rpc_client.send_email(
            user_ids, mail_subject, mail_message, countdown, eta, ignore_err)

    def send_sms(self, user_ids, sms_code, sms_params, countdown=None, eta=None, ignore_err=True):
        """
        发送短信
            @parms user_ids: 接收短信的用户id (可以为list)
            @parms sms_code: sms_code (在阿里短信后台配置)
            @parms sms_params: 模板参数 (必须与配置的模板匹配)
            @parms countdown: 是否延迟发送, 延时几秒, 例如: countdown=5
            @parms eta: 是否延迟发送, 延时几秒, 例如: eta=datetime.datetime.utcnow() + datetime.timedelta(seconds=5) 注意时间必须为 utc时间
            @parms ignore_err: 是否忽略错误

        例:
            InnerServices.MSG().send_sms([57], 'SMS_177552091', {'username': 'tester','password': 'pwd12345'}, 3)
        """
        return self.rpc_client.send_sms(user_ids, sms_code, sms_params, countdown, eta, ignore_err)



# -*- coding: utf-8 -*-

import traceback
from django.http import JsonResponse
from django.views.decorators.csrf import csrf_exempt
from base64 import b64encode
from functools import wraps
from ywkd_tools.auth import Auth


@Auth.inner
def token_required(view_func):
    """检查用户是否有某权限"""

    @csrf_exempt
    @wraps(view_func)
    def _wrapped_view(request, *args, **kwargs):
        basic_auth = request.META.get('HTTP_AUTHORIZATION')

        # 兼容老接口认证规则
        if not basic_auth and type(request.POST) != type([]):
            user_id = request.POST.get('user', request.GET.get('user'))
            token = request.POST.get('token', request.GET.get('token'))
            basic_auth = 'basic ' + b64encode('{}:{}'.format(user_id, token).encode('utf-8')).decode('utf-8')
            request.META['HTTP_AUTHORIZATION'] = basic_auth

        user = None
        try:
            user = Auth.RSTFBasciAuthentication().authenticate(request)[0]
        except Exception as e:
            # 兼容老接口认证规则
            traceback.print_exc()
            result = {"data": []}
            result.update({"code": 500102, "codemsg": u"token解析错误/权限限制"})
            return JsonResponse(result)

        request.user = user
        return view_func(request, *args, **kwargs)

    return _wrapped_view



from ywkd_tools.inner_service_apis.inner_service_apis import InnerServices, InnerService
from ywkd_tools.inner_service_apis.errors import InnerAPIError


@InnerServices.inner
class Cperm(InnerService):
    """用户中心"""

    url = '/cperm/rpc/'

    def get_user(self, user_id):
        """
        获取用户详情
            @param: user_id 用户id
            @out: {     // 返回内容参考: http://yapi.yuwangkedao.com/project/251/interface/api/6341
                id: [用户id],
                user_type: 用户类型,
                avatar: 头像,
                belong_participant: {   // 所属公司
                    id: [公司id],
                    participant_type: [公司类型]
                    ...
                },
                ...
            }
        """
        return self.rpc_client.get_user(user_id)

    def check_api_operate_perms(self, user_id, token, service_name, api_name, operator_type):
        """
        检测当前token是否可以对当前 api 进行操作
            @param: user_id 用户id
            @param: token 用户token
            @param: service_name 服务名称 例如"KTV"
            @param: api_name api名
            @param: operator_type 操作类型 比如: GET, POST, PUT

            @out {"code": [返回码], "codemsg": [消息内容]}
                成功: {"code": 0, "codemsg": u"请求成功"}
                失败: {"code": 500104, "codemsg": u"用户被禁用或token值不对"}
                      {"code": 500102, "codemsg": u"用户无权限"}
                      {"code": 500016, "codemsg": u"用户被禁用/已删除"}
        """
        return self.rpc_client.check_api_operate_perms(
            user_id, token, service_name, api_name, operator_type)

    def filter_users(self, id__in=None, user_type__in=None, phone__in=None, nickname__contains=None,
                is_active=None, is_all_area=None, belong_participant_id__in=None,
                group_codes__in=None, permission_codes__in=None, page_index=1, page_size=100):
        """
        过滤用户:
            例:
                # 过滤 角色code 为 CEO,CTO 的用户
                filter_users(None, None, None, None, None, None, None, ['ceo', 'cto'])

            @out:
                {
                    "total": 2,     # 总数
                    "page_index": 1,    # 当前页 起始页 1
                    "page_size": 100,   # 每页大小
                    "data": [{
                        "id": 57,       # user_id
                        "user_type": "employee",    # 用户类型
                        "phone": "13644092",        # 手机号
                        "nickname": "昵称A",        # 昵称
                        "gender": 1,                # 性别
                        "is_all_area": False,       # 是否全国
                        "area_code_list": [],       # 区域codes
                        "serial_number": "000018",  # 序列号
                        "belong_participant": None, # 所属公司
                        "group_codes": ["ceo"]      # 所属组 codes
                    }, {
                        "id": 21,
                        "user_type": "agentibus",
                        "phone": "13644099810",
                        "nickname": "ABC",
                        "gender": 1,
                        "is_all_area": True,
                        "area_code_list": ["00320200", "00330100"],
                        "serial_number": "110007",
                        "belong_participant": {
                            "id": 1005,
                            "participant_type": "agentibus",
                            "area_code_list": ["00120000", "00120100", "00120101"],
                            "address": "西溪壹号",
                            "contact": "联系人A",
                            "telephone": "0411-84620027",
                            "comment": "备注B",
                            "serial_number": "DLS0005"
                        },
                        "group_codes": ["ceo"]
                    }]
                }
        """
        return self.rpc_client.filter_users(
                id__in, user_type__in, phone__in, nickname__contains,
                is_active, is_all_area, belong_participant_id__in,
                group_codes__in, permission_codes__in, page_index, page_size)



import logging
import log_request_id

logger = logging.getLogger('ywkd_tools')
logger.setLevel(logging.INFO)

hd = logging.StreamHandler()
fm = logging.Formatter('{"asctime": "%(asctime)s", "levelname": "%(levelname)s", "location": "%(module)s.%(funcName)s", "path":"%(pathname)s:%(lineno)s", "request_id": "%(request_id)s", "message": "%(message)s"}')
hd.setFormatter(fm)
hd.addFilter(log_request_id.filters.RequestIDFilter())
logger.addHandler(hd)



# -*- coding: utf-8 -*-

from .inner_service_apis import InnerServices
from .errors import InnerAPIError
from .cperm import *
from .msg import *
from .participant import *

__all__ = [InnerServices, InnerAPIError]



# -*- coding: utf-8 -*-

import os
import sys
import setuptools
from setuptools import setup


path, script = os.path.split(sys.argv[0])
os.chdir(os.path.abspath(path))


with open('README.md') as f:
    long_description = f.read()

setuptools.setup(
    name='ywkd_tools',
    version='1.0.16',
    description='娱网科道内部python 公用包',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='ywkd',
    author_email='',
    url='',
    packages=setuptools.find_packages(),
    install_requires=[
        'django', 'djangorestframework >= 3, < 4', 'django-log-request-id'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.4',
)
